#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <util/delay.h>
#include "lcd-routines.h"

#define BUTTON_ADC 0
#define LASER_ADC 1
#define MIC_ADC 2

/* Button related variables */
char msgs[5][15] = {"Right Key OK ", 
                    "Up Key OK    ", 
                    "Down Key OK  ", 
                    "Left Key OK  ", 
                    "Select Key OK" };

int adc_key_val[5] ={30, 150, 360, 535, 760 };
int NUM_KEYS = 5;
int adc_key_in;
int key=-1;

/* FSM related variables */
int _NUMSTATES = 4;
enum statevals{
	INTERVALLOMETER=0,
	LASER,
	MIC
};
int state = -1;

/* Buffer for LCD messages */
char buffer[20];

/* Mode related variables */
int intval = 0;
int running = 0;
int initial_value = 0;
int user_intval = 10;


/* 
 * Read analog to digital konverter
 * Takes pin number as input
 */
void switch_adc(int input){
	ADCSRB |= (1<<ACME); // enable multiplexing
	switch(input){
		case BUTTON_ADC:
			ADMUX = (0<<REFS1 | 1<<REFS0 | 0<<ADLAR | 0<<MUX3 | 0<<MUX2 | 0<<MUX1 | 0<<MUX0);
			// prescaler: 128
			ADCSRA = (1<<ADEN | 0<<ADSC | 0<<ADIF | 1<<ADPS2 | 1<<ADPS1 | 1<<ADPS0);
			break;
		case LASER_ADC:
			ADMUX = (0<<REFS1 | 1<<REFS0 | 0<<ADLAR | 0<<MUX3 | 0<<MUX2 | 0<<MUX1 | 1<<MUX0);
			// prescaler: 16
			ADCSRA = (1<<ADEN | 0<<ADSC | 0<<ADIF | 1<<ADPS2 | 0<<ADPS1 | 0<<ADPS0);
			break;
		case MIC_ADC:
			ADMUX = (0<<REFS1 | 1<<REFS0 | 0<<ADLAR | 0<<MUX3 | 0<<MUX2 | 1<<MUX1 | 0<<MUX0);
			// prescaler: 16
			ADCSRA = (1<<ADEN | 0<<ADSC | 0<<ADIF | 0<<ADPS2 | 0<<ADPS1 | 0<<ADPS0);
			break;
		case 3:
			ADMUX = (0<<REFS1 | 1<<REFS0 | 0<<ADLAR | 0<<MUX3 | 0<<MUX2 | 1<<MUX1 | 1<<MUX0);
			break;
		case 4:
			ADMUX = (0<<REFS1 | 1<<REFS0 | 0<<ADLAR | 0<<MUX3 | 1<<MUX2 | 0<<MUX1 | 0<<MUX0);
			break;
		case 5:
			ADMUX = (0<<REFS1 | 1<<REFS0 | 0<<ADLAR | 0<<MUX3 | 1<<MUX2 | 0<<MUX1 | 1<<MUX0);
			break;
	}
	_delay_ms(50);
	ADCSRB &= ~(1<<ACME); // disable multiplexing

	_delay_ms(10);
	/* dummy reading */
	ADCSRA = (1<<ADEN | 1<<ADSC | 0<<ADIF | 1<<ADPS2 | 1<<ADPS1 | 1<<ADPS0);
	while(ADCSRA & (1<<ADSC)){
			; // Wait until conversion is finished
	}
	/* read result */
	int result = ADCL;
	result += (ADCH<<8);
}

int read_adc(){
	ADCSRA = (1<<ADEN | 1<<ADSC | 0<<ADIF | 1<<ADPS2 | 1<<ADPS1 | 1<<ADPS0);
	ADCSRA |= (1<<ADSC);
	while(ADCSRA & (1<<ADSC)){
		;
	}
	int result = ADCL;
	result += (ADCH<<8);

	return result;
}


/* Display current state */
void show_state(){
	lcd_set_cursor(0,1);
	switch(state){
		case 0:
			lcd_string("Intervallometer ");
			lcd_set_cursor(0,2);
			lcd_string("                ");
			lcd_set_cursor(5,2);
			lcd_string("VAL:");
			itoa(user_intval, buffer, 10);
			lcd_string(buffer);
			break;
		case 1:
			lcd_string("Laser Trigger   ");
			lcd_set_cursor(0,2);
			lcd_string("                ");
			break;
		case 2:
			lcd_string("Mic Trigger     ");
			lcd_string("                ");
			break;
	}
}

/* set state depending on key input */
void set_state(int key){
	switch(key){
		case 0: // right key
			if(running==0){
				if(state == INTERVALLOMETER){
					user_intval++;
					lcd_set_cursor(5,2);
					lcd_string("VAL:");
					itoa(user_intval, buffer, 10);
					lcd_string(buffer);
				}
			}
			break;
		case 1: // up key
			if(running==0){
				state++;
				state = state % _NUMSTATES;
			}
			break;
		case 2: //down key
			if(running==0){
				state--;
				if(state < 0){
					state += _NUMSTATES;
				}
				state = state % _NUMSTATES;
			}
			break;
		case 3: // left key
			if(running==0){
				if(state == INTERVALLOMETER){
					user_intval--;
					if(user_intval <= 1){
						user_intval = 1;
					}
					lcd_set_cursor(5,2);
					lcd_string("VAL:");
					itoa(user_intval, buffer, 10);
					lcd_string(buffer);
				}
			}
			break;
		case 4: // select key
			lcd_set_cursor(0,2);
			lcd_string("                ");
			if(running == 0){
				running = 1;
				switch(state){
					case LASER:
						/* read a value and set this as compare-value */
						switch_adc(LASER_ADC);
						initial_value = read_adc();
						break;
					case MIC:
						/* read a value and set this as compare-value */
						switch_adc(MIC_ADC);
						initial_value = read_adc();
						break;
				}
			}else if(running==1){
				running = 0;
				intval = 0; // reset time-counter
			}
			break;
	}
	show_state();
	/* delay to prevent one key press from causing multiple actions */
	_delay_ms(700);
}

/*
 * Reads the button-adc and
 * matches it to a key number
 */
int get_key(){
	int k;
	switch_adc(BUTTON_ADC);
	int result = read_adc();
	/* key mapping */
	for (k = 0; k < NUM_KEYS; k++){
		if (result < adc_key_val[k]){
			return k;
		}
	}
	if (k >= NUM_KEYS) k = -1;     // No valid key pressed
	return k;
}

/* triggers the camera */
void trigger(){
	/* connect all three cam-pins to GND */
	PORTD &= ~(1 << 1);
	PORTD &= ~(1 << 2);

	if(state == MIC){
		_delay_ms(20);
	}else{
		_delay_ms(50);
	}

	/* set high again, "untrigger" */
	PORTD |= (1 << 1);
	PORTD |= (1 << 2);
}

/* inits ports and lcd */
void setup() {
	lcd_init();
	lcd_string(" MEGAVALLOMETER ");
	lcd_set_cursor(0,2);
	lcd_string("UP/DOWN FOR MODE");
	DDRD = 0xff;
	DDRC = 0x00;

	PORTD &= ~(1 << 0); // connect CAM-GND to our GND
	PORTD |= (1 << 1); // other two pins high to prevent triggering
	PORTD |= (1 << 2);
}

/*
 * main loop
 * -> chech for button input
 * -> fsm for the different modes
 */
void loop() {
	int result = 0;

	/* if not started, only display sensor readings */
	if(running == 0){
		int key = get_key();
		if(key != -1){
			set_state(key);
		}
		switch(state){
			case INTERVALLOMETER:
				break;
			case LASER:
				switch_adc(LASER_ADC);
				result = read_adc();
				lcd_set_cursor(0,2);
				lcd_string("                ");
				lcd_set_cursor(0,2);
				itoa(result, buffer, 10);
				lcd_string(buffer);
				break;
			 case MIC:
				switch_adc(MIC_ADC);
				result = read_adc();
				lcd_set_cursor(0,2);
				lcd_string("                ");
				lcd_set_cursor(0,2);
				lcd_set_cursor(0,2);
				itoa(result, buffer, 10);
				lcd_string(buffer);

				/* normalize values to 16 to display as bar graph */
				/*
				equalizer = result / 1024.0 * 16.0;
				for(count = 0; count < equalizer; count++){
					lcd_string("#");
				}
				*/
				break;
		}
		_delay_ms(10);
	}

	/* if started, execute actions */
	if(running == 1){
		switch(state){
			case INTERVALLOMETER:
				if(intval >= user_intval*10){
					/* time is up, trigger */
					intval = 0;
					trigger();
				}else{
					/* increase counter and display */
					lcd_set_cursor(0,2);
					lcd_string("                ");
					lcd_set_cursor(0,2);
					itoa((intval/10), buffer, 10);
					lcd_string(buffer);
					lcd_set_cursor(5,2);
					lcd_string("VAL:");
					itoa(user_intval,buffer, 10);
					lcd_string(buffer);
					_delay_ms(100); //1/10th second
					intval++;
				}
				break;
			case LASER:
				/* initial value is reading with laser on */
				/* so to trigger, result has to be significantly lower */
				result = read_adc();
				if(result < initial_value - 32){
					trigger();
					lcd_set_cursor(0,2);
					lcd_string("TRIGGER!        ");
					/* 
					itoa(initial_value, buffer, 10);
					lcd_string(buffer);
					lcd_set_cursor(8,2);
					itoa(result, buffer, 10);
					lcd_string(buffer);
					*/
					_delay_ms(1000);
					lcd_set_cursor(0,2);
					lcd_string("                ");
				}
				break;
			 case MIC:
				/* initial value is environment noise */
				/* so to trigger, result has to be significantly higher */
				//switch_adc(MIC_ADC);
				result = read_adc();
				if(result > initial_value + 32){
					trigger();
					lcd_set_cursor(0,2);
					lcd_string("TRIGGER!        ");
					_delay_ms(1000);
					lcd_set_cursor(0,2);
					lcd_string("                ");
				}
				break;
		}
	}
}

int main(){
	setup();
	for(;;){
		loop();
	}

	return 0;
}
